import uuid
import json
from .room import Room


class Connection():
    ws = None
    hub = None
    user_id = None

    def __init__(self, hub, ws):
        self.ws = ws
        self.hub = hub
        self.subscriptions = {}
        self.allowed_rooms = []
        self.uuid = uuid.uuid4()

        # Nasty hack
        try:
            ws.connection = self
        except AttributeError:
            pass

    def parse_user(self, user_data):
        self.user_id = user_data['id']

    def handle_auth_success(self, data):
        self.user_id = data['user']['id']
        self.allowed_rooms = data['allowed_rooms']

        self.send({'allowed_rooms': self.allowed_rooms})

    def handle(self, message):
        if message == 'ping':
            self.ws.send('pong')
            return

        m = json.loads(message)

        requestId = m.get('requestId', None)

        if m['type'] == 'subscribe':
            return self.handle_subscribe(m)

        if m['type'] == 'unsubscribe':
            return self.handle_unsubscribe(m)

        self.send({
            'requestId': requestId,
            'code': 'error',
            'message': 'message-type-not-allowed',
        })

    # If all keys match for a certain room,
    def is_room_allowed(self, room_dict):
        def room_matches(rd, ar):
            if len(rd.keys()) != len(ar.keys()):
                return False
            for ar_key in ar.keys():
                if ar_key not in rd:
                    return False
                if ar[ar_key] == '*':
                    continue
                if rd[ar_key] != ar[ar_key]:
                    return False

            return True

        for ar in self.allowed_rooms:
            if room_matches(room_dict, ar):
                return True

        return False

    def handle_subscribe(self, m):
        room_dict = m.get('room', None)
        room_hash = Room.hash_dict(room_dict)

        if not self.is_room_allowed(room_dict):
            self.send({
                'requestId': m['requestId'],
                'code': 'error',
                'message': 'room-not-found',
            })
            return

        sub = self.hub.subscribe(self, m, room_hash)
        self.subscriptions[m['requestId']] = sub
        self.send({
            'requestId': m['requestId'],
            'code': 'success',
        })

    def handle_unsubscribe(self, m):
        reqId = m['requestId']
        if reqId not in self.subscriptions:
            self.send({
                'requestId': m['requestId'],
                'code': 'error',
                'message': 'not-subscribed',
            })
            return

        sub = self.subscriptions[reqId]
        sub.stop()

        self.subscriptions.pop(reqId)
        self.send({
            'requestId': reqId,
            'code': 'success',
        })

    def unsubscribe_all(self):
        for room in [sub.room for sub in self.subscriptions.values()]:
            room.remove_connection(self)

        self.subscriptions = {}

    def send(self, message):
        if self.ws.closed:
            return

        self.ws.send(json.dumps(message))



import json
from .testapp.app import app
from high_templar.test import TestCase, Client, MockWebSocket, room_car, room_car_reverse, MockResponse

subscribe_car = {
    'requestId': 'a',
    'type': 'subscribe',
    'room': room_car,
}
subscribe_car_reverse = {
    'requestId': 'a',
    'type': 'subscribe',
    'room': room_car_reverse,
}
subscribe_weird = {
    'requestId': 'b',
    'type': 'subscribe',
    'room': {'foo': 'bar'},
}


class TestSubscribe(TestCase):
    def setUp(self):
        self.client = Client(app)

    def test_subscribe_success(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(json.dumps(subscribe_car))
        self.client.open_connection(ws)

        self.assertEqual({'requestId': 'a', 'code': 'success'}, json.loads(ws.outgoing_messages[1]))

    def test_room_key_order_unimportant(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(json.dumps(subscribe_car_reverse))
        self.client.open_connection(ws)

        self.assertEqual({'requestId': 'a', 'code': 'success'}, json.loads(ws.outgoing_messages[1]))

    def test_subscribe_unallowed_room(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(json.dumps(subscribe_weird))
        self.client.open_connection(ws)

        self.assertEqual({'requestId': 'b', 'code': 'error', 'message': 'room-not-found'}, json.loads(ws.outgoing_messages[1]))


class TestAllowedRoomMatch(TestCase):
    def setUp(self):
        self.client = Client(app)

    def test_wildcard(self):
        allowed_car_all = {'target': 'car', 'car': '*'}

        def allowed_car_wildcard(request, **kwargs):
            return MockResponse({
                'user': {'id': 1},
                'allowed_rooms': [allowed_car_all],
            }, 200)

        self.client.set_mock_api(allowed_car_wildcard)

        ws = MockWebSocket()
        ws.mock_incoming_message(json.dumps(subscribe_car))
        self.client.open_connection(ws)

        self.assertEqual({'requestId': 'a', 'code': 'success'}, json.loads(ws.outgoing_messages[1]))

    def test_empty_dict(self):
        def empty_dict_allowed(request, **kwargs):
            return MockResponse({
                'user': {'id': 1},
                'allowed_rooms': [{}],
            }, 200)

        self.client.set_mock_api(empty_dict_allowed)

        ws = MockWebSocket()
        ws.mock_incoming_message(json.dumps(subscribe_car))
        self.client.open_connection(ws)

        self.assertEqual(2, len(ws.outgoing_messages))
        self.assertEqual({'requestId': 'a', 'code': 'error', 'message': 'room-not-found'}, json.loads(ws.outgoing_messages[1]))

    def test_mismatched_key(self):
        def empty_dict_allowed(request, **kwargs):
            return MockResponse({
                'user': {'id': 1},
                'allowed_rooms': [{
                    'target': 'car',
                    'foo': 1,
                }],
            }, 200)

        self.client.set_mock_api(empty_dict_allowed)

        ws = MockWebSocket()
        ws.mock_incoming_message(json.dumps(subscribe_car))
        self.client.open_connection(ws)

        self.assertEqual(2, len(ws.outgoing_messages))
        self.assertEqual({'requestId': 'a', 'code': 'error', 'message': 'room-not-found'}, json.loads(ws.outgoing_messages[1]))



import json


class WebSocketClosedError(Exception):
    pass


class Subscription:
    '''
    A subscription of a connection to a room
    '''

    def __init__(self, connection, room, request):
        self.connection = connection
        self.room = room
        self.requestId = request['requestId']
        self.scope = request.get('scope', {})

    def publish(self, data):
        if self.connection.ws.closed:
            raise WebSocketClosedError

        self.connection.ws.send(json.dumps({
            'requestId': self.requestId,
            'type': 'publish',
            'data': data,
        }))

    def stop(self):
        self.room.remove_subscription(self)


class Room:
    '''
    A collection of subscriptions which have access to a scoped permission.

    A permission would be: See actions
    A scoped permission would be: See your own actions

    Then there would "exist" a room for every single user with view_action.own scoped permission
    "exist" between quotation marks, because high-templar only creates a room when someone actually subscribes to it.

    A room keeps track of all connections with the same scoped permission.
    In the case a view_action.own, it would have a maximum of 1 connection.
    '''

    def __init__(self, hash, hub):
        self.hash = hash
        self.hub = hub
        self.subscriptions = []

    def subscribe(self, connection, request):
        sub = Subscription(connection, self, request)
        self.subscriptions.append(sub)
        return sub

    # Don't just unsubscribe: remove all subscriptions for a connection
    def remove_connection(self, connection):
        # Clone self.subscriptions because they may be removed when closed
        for s in list(self.subscriptions):
            if s.connection == connection:
                self.subscriptions.remove(s)

        self.close_if_empty()

    def remove_subscription(self, sub):
        self.subscriptions.remove(sub)

        self.close_if_empty()

    def close_if_empty(self):
        if not self.subscriptions:
            self.close()

    def close(self):
        self.hub.close_room(self)

    # Returns a list of closed connections
    def publish(self, data):
        closed_connections = []
        for s in self.subscriptions:
            try:
                s.publish(data)
            except WebSocketClosedError:
                closed_connections.append(s.connection)

        return closed_connections

    @property
    def connections(self):
        return [s.connection for s in self.subscriptions]

    # Room names are dict.
    # Every keyval pair is a scope of the permission.
    # IE:
    # { 'zoo': 1, 'animal_type': 'penguin'}
    # Is a room where publishes about the penguins in zoo 1 appear.
    #
    # The backend needs to dictate which rooms a user can be in,
    # and the frontend needs to programmatically define the room to subscribe to.
    #
    # We hash the dicts so that
    # {'foo': True, 'bar': True} is the same room as {'bar': True, 'foo': True}
    # and so we can still check if a room exists
    @staticmethod
    def hash_dict(room_dict):
        return json.dumps(room_dict, sort_keys=True)



import json
from .testapp.app import app
from high_templar.test import TestCase, Client, MockWebSocket, room_car
from greenlet import greenlet

subscribe_car = {
    'requestId': 'car',
    'type': 'subscribe',
    'room': room_car,
}

subscribe_car_2 = {
    'requestId': 'car_2',
    'type': 'subscribe',
    'room': room_car,
}

unsubscribe_car = {
    'requestId': 'car',
    'type': 'unsubscribe',
}

unsubscribe_car_2 = {
    'requestId': 'car_2',
    'type': 'unsubscribe',
}


class TestUnsubscribe(TestCase):
    def setUp(self):
        self.client = Client(app)

    def test_no_more_publish(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(json.dumps(subscribe_car))

        g = greenlet(self.client.open_connection)
        g.switch(ws)

        self.assertEqual(2, len(ws.outgoing_messages))
        self.assertEqual({'requestId': 'car', 'code': 'success'}, json.loads(ws.outgoing_messages[1]))

        self.trigger({
                'rooms': [room_car],
                'data': [{'id': 1}]
            })

        g.switch()
        self.assertEqual(3, len(ws.outgoing_messages))
        self.assertEqual({
            'requestId': 'car',
            'type': 'publish',
            'data': [{'id': 1}]
            }, json.loads(ws.outgoing_messages[2]))

        ws.mock_incoming_message(json.dumps(unsubscribe_car))

        g.switch()
        self.assertEqual(4, len(ws.outgoing_messages))
        self.assertEqual({'requestId': 'car', 'code': 'success'}, json.loads(ws.outgoing_messages[3]))

        self.trigger({
                'rooms': [room_car],
                'data': [{'id': 1}]
            })

        g.switch()
        self.assertEqual(4, len(ws.outgoing_messages))

    def test_sub_removed_from_room(self):
        ws1 = MockWebSocket()
        ws1.mock_incoming_message(json.dumps(subscribe_car))
        ws2 = MockWebSocket()
        ws2.mock_incoming_message(json.dumps(subscribe_car_2))

        g1 = greenlet(self.client.open_connection)
        g1.switch(ws1)
        g2 = greenlet(self.client.open_connection)
        g2.switch(ws2)

        room = self.getHubRoomByDict(room_car)
        c1 = room.connections[0]
        c2 = room.connections[1]

        self.assertHubRoomsEqual([room_car])
        self.assertEqual([c1, c2], room.connections)
        self.assertEqual([room], [s.room for s in c1.subscriptions.values()])
        self.assertEqual([room], [s.room for s in c2.subscriptions.values()])

        ws2.mock_incoming_message(json.dumps(unsubscribe_car_2))
        g2.switch()

        self.assertHubRoomsEqual([room_car])
        self.assertEqual([c1], room.connections)
        self.assertEqual([room], [s.room for s in c1.subscriptions.values()])
        self.assertEqual(0, len(c2.subscriptions))

        ws1.mock_incoming_message(json.dumps(unsubscribe_car))
        g1.switch()

        self.assertHubRoomsEqual([])
        self.assertEqual(0, len(c1.subscriptions))
        self.assertEqual(0, len(c2.subscriptions))

    def test_cannot_unsubscribe_when_not_subscribed(self):
        ws = MockWebSocket()
        ws.mock_incoming_message(json.dumps(unsubscribe_car))

        self.client.open_connection(ws)
        self.assertEqual(2, len(ws.outgoing_messages))
        self.assertEqual({'requestId': 'car', 'code': 'error', 'message': 'not-subscribed'}, json.loads(ws.outgoing_messages[1]))
